/* File yasep/core-js/yasep_disasm.js
created 2006-09-04 by whygee@f-cpu.org
2006-11-10 (corrected an edition glitch ???)
2007-04-04 : update for the Dx pseudo-instructions
2007-04-09 : separation of forms and flags
2007-04-17 : another major update
2007-11-05 : rebranding to YASEP, support of the flags IMM16_5LSB and IMM16_8LSB
2009-07-30 : Y and others
2009-08-10 : rebuild !
2009-08-15 : RIR=>IRR, +decode_extended
2009-08-18 : factoring ? trying a new approach where the 5 main forms are separated,
              and 5 functions try to infer the aliases and short versions
2009-08-23 : working again
2011-09-09 : new object model, add_message part of the context

This file does the disassembly (decoding) of encoded instructions.
It transforms a number into a string and a size, for interactive debug and testing.

dependency:
 yasep/JScore/txt_int.js,
///////////// (yasep/JSgui/yasep_messages.js+handling [define an empty function if undesired]),
 yasep/JScore/yasep_forms.js/flags/groups/fields
 yasep/JScore/yasep_opcodes.js+children (group_asu.js etc.)
*/

// This is the constructor of the object that is being passed to and from the assembler
function DISASM_context(add_message) {
  // input
  if (typeof add_message=="function")
    this.add_message = add_message;
  else
    this.add_message = function() {};
  this.YASEP_TYPE=0; // YASEP16/YASEP32 ?
}


// Just a big function with several local sub-functions
function Yasep_DISASM(input, t){

  function decode_short(t,f,g){
// decode the Imm4 forms first
    var i4=t.si4;
    if (!(g & Y.IgnoreImmSign) && (i4 >= 8))
      i4 -= 16;

    if (t.input & Y.FIELD_EXTFORM){
      // aliases :
      if (t.si4==0){
        if (t.name=="SUB"){
          t.name="NEG";
          t.instruction=" "+Y.registers_names[t.snd];
          return 'FORM_R'
        }
        if (t.name=="ORN"){
          t.name="NOT";
          t.instruction=" "+Y.registers_names[t.snd];
          return 'FORM_R'
        }
      }
      if (f & Y.FORM_iR){
        t.instruction=" "+i4+" "+Y.registers_names[t.snd];
        return 'FORM_iR'
      }
      if(f & Y.FORM_Ri){
        t.instruction=" "+Y.registers_names[t.snd]+" "+i4;
        return 'FORM_Ri'
      }
    }
    if (f & Y.FORM_i){ // CRIT
      t.instruction=" "+i4;
      if (t.snd != 0)
        t.add_message('yellow', 'warning : garbage found in SND');
      return 'FORM_i'
    }

    // some ALONE are already filtered out but not the aliases :
    if (t.input ==0) {
      t.name="NOP";
      return 'FORM_ALONE'
    }

    // RR :
    t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.snd];
    return 'FORM_RR'
  };

  //////////////////////////// IRR ////////////////////////////
  function decode_long(t,f,g){
    var l=(t.input>>16)&65535;
    var s=16;
    if ((g & Y.IMM20) && (t.YASEP_TYPE!=16)) {
      s=20;
      l |= t.snd<<16;
    }

    if (!(g & Y.IgnoreImmSign) && (l >= (1<<(s-1))))
      l -= (1<<s);

    if (f & Y.IGNORE_IMM16){ // does it ignore the Imm16 field ?
      t.form='FORM_ALONE';
      if (l!=0)
        t.add_message('yellow', 'warning : garbage found in IMM16');
      if (t.snd != 0)
        t.add_message('yellow', 'warning : garbage found in SND');
      if (t.si4 != 0)
        t.add_message('yellow', 'warning : garbage found in SI4');
    }

    if (f&Y.FORM_RI) { // PUT
      if (t.si4 != 0)
        t.add_message('yellow', 'warning : garbage found in SI4');
      t.instruction=" "+Y.registers_names[t.snd]+" "+l;
      return 'FORM_RI'
    }   

    if (f&Y.FORM_IR) {
      if (g&Y.ALIAS_IRR){
        if (t.snd==t.si4) { // condensed writing
          t.instruction=" "+l+" "+Y.registers_names[t.snd]; // either si4 or snd, they are equal...
          return 'FORM_IR'
        }
      }
      else {
        if (g & Y.SWAP_IR){ // MOV, GET
          if ((s!=20) && (t.snd != 0))
            t.add_message('yellow', 'warning : garbage found in SND');
          t.instruction=" "+l+" "+Y.registers_names[t.si4];
        }
        else { // CMPU, CMPS
          if (t.si4 != 0)
            t.add_message('yellow', 'warning : garbage found in SI4');
          t.instruction=" "+l+" "+Y.registers_names[t.snd];
        }
        return 'FORM_IR'
      }
    }
    // default
    t.instruction=" "+l+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.si4];
    return 'FORM_IRR'
  }

  function decode_extended(t,f,g){
  // decode the Imm4 forms first
    if (t.input & Y.FIELD_EXTIMM4) {
      var i4=t.si4;
      if (!(g & Y.IgnoreImmSign) && (i4 >= 8))
        i4 -= 16;

      // aliases :
      if ((t.si4==0)&&(t.snd==t.dst3)){
        if (t.name=="SUB"){
          t.name="NEG";
          t.instruction=" "+Y.registers_names[t.snd];
          if (t.snd == t.dst3)
            return 'FORM_R'
          t.instruction=" "+Y.registers_names[t.dst3];
          return 'FORM_RR'
        }
        if (t.name=="ORN"){
          t.name="NOT";
          t.instruction=" "+Y.registers_names[t.snd];
          if (t.snd == t.dst3)
            return 'FORM_R'
          t.instruction=" "+Y.registers_names[t.dst3];
         return 'FORM_RR'
        }
      }
      if ((f & Y.FORM_iR)&&(t.snd == t.dst3)) {
        t.instruction=" "+i4+" "+Y.registers_names[t.snd];
        return 'FORM_iR'
      }
      if(f & Y.FORM_Ri){ // PUT
        t.instruction=" "+Y.registers_names[t.snd]+" "+i4;
        return 'FORM_Ri'
      }
      // default
      if (f & Y.FORM_iRR) {

        if (t.snd == t.dst3) {
          t.instruction=" "+i4+" "+Y.registers_names[t.snd];
          return 'FORM_iR'
        }

        t.instruction=" "+i4+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.dst3];
        return 'FORM_iRR'
      }
    }
    else {
      if ((f & Y.FORM_RR)&&(t.snd == t.dst3)) {
        t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.dst3];
        return 'FORM_RR'
      }
      if (f & Y.FORM_RRR) {
        t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.dst3];
        return 'FORM_RRR'
      }
    }
    add_message('red', 'Error : unknown or unhandled form ???');
  }

  // TODO: MUST BE REDESIGNED
  function decode_condition(t){
    var cond=(t.input & Y.FIELD_EXTCOND)>>Y.OFFSET_EXTCOND;
    if ((cond!=0) || (t.src4!=0)) { // exclude the ALWAYS condition
      var c=Y.tableCondition[cond]; // basename of the condition
      var o=Y.conditionTable[c]; // the condition object
      var n=Y.registers_names[t.src4]; // name of the register
      d=o[n]; // check if the register is associated with an alias
      if (typeof d=='string')
        t.instruction+=" "+d;
      else {
        t.instruction+=" "+c;
        if (o.reg != t.src4) // always, never
          t.instruction+=" "+n;
      }
    }   
  }


////////////////////////////////////////////////////
// Start of the decoding function

// t.decode_instruction=function(input) {
  t.form=null; // indicates the error if still null after return
  t.prefix=input & (Y.FIELD_IMMREG|Y.FIELD_EXTFORM);
  t.opcode=input &  Y.FIELD_OPCODE;
  t.input=input;
  t.instruction='';
  t.size_bytes=0;
  t.snd=(input & Y.FIELD_SND)>>Y.OFFSET_SND;
  t.si4=(input & Y.FIELD_SI4)>>Y.OFFSET_SI4;
  t.name=null;

  if (typeof input != 'number') {
    add_message('red', 'Error : wrong input format');
    return
  }

  // opcode lookup
  t.O=Y.opcode_table[t.opcode];

  if (typeof t.O != 'object') {
    if (input & Y.FIELD_IMMREG) {
      // output a DW
      t.add_message('red', 'Error : unknown opcode "'+int2hex(opcode)+'" (assuming a long instruction)');
      t.instruction = 'DW '+ int2hex(input_instruction,7,7);
      t.size_byte=4;
    }
    else {
      // output a DH
      t.add_message('red', 'Error : unknown opcode "'+int2hex(opcode)+'" (assuming a short instruction)');
      if (input & 0xFFFF0000)
         t.add_message('orange', 'Warning : garbage in the MSB');
      t.instruction = 'DH '+ int2hex(input_instruction & F4);
      t.size_byte=2;
    }
    return
  }

  t.name=t.O.name;
  var f=t.O.forms;
  var g=t.O.flags;

  // the size is straightforward to find :
  if (input & Y.FIELD_IMMREG) {
    // here we process a long instruction 
    t.size_bytes=4;

    // extended forms :
    if (input & Y.FIELD_EXTFORM) {
      t.dst3=(input>>Y.OFFSET_EXTDST3)&15;
      t.src4=(input>>Y.OFFSET_EXTSRC4)&15;
      t.form=decode_extended(t,f,g);
      decode_condition(t);
    }
    else { // Long Immediate instructions : FORM_RI, FORM_IR, FORM_IRR
      // test the garbage :
      if((input & 0xFF000000)&&(g&Y.IMM16_8LSB)
      || (input & 0xFFE00000)&&(g&Y.IMM16_5LSB))
        t.add_message('yellow', 'warning : garbage found in IMM16');
      t.form=decode_long(t,f,g);
    }
  }
  else { // Short forms
    t.size_bytes=2;
    if (f & Y.FORM_ALONE) {
      t.form='FORM_ALONE';
      if (t.snd != 0)
        add_message('yellow', 'warning : garbage found in SND');
      if (t.si4 != 0)
        add_message('yellow', 'warning : garbage found in SI4');
    }
    else
      t.form=decode_short(t,f,g);
  }

  // output
  if (t.form!=null){
    t.instruction=t.name+t.instruction;
    t.add_message('blue',' OK: opcode='+Y.decorate_opcode(t.opcode)+' instruction="<b>'+t.instruction+'</b>" ('+t.form+')');
  }
}
